/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 *  
 *  2/27/00:3am, random plans by jer
 *  
 *  ok based on gprof, we really need some innovation here... my thoughs are this:
 *  
 *  most things are strings, so have a string-based true-blue garbage collector
 *  one big global hash containing all the strings created by any pstrdup, returning const char *
 *  a refcount on each string block
 *  when a pool is freed, it moves down the refcount
 *  garbage collector collects pools on the free stack, and runs through the hash for unused strings
 *  j_strcmp can check for == (if they are both from a pstrdup)
 *  
 *  let's see... this would change:
 *  pstrdup: do a hash lookup, success=return, fail=pmalloc & hash put
 *  pool_free: 
 *  
 *  
 *  
 *  
 *  
 */

#include "libxode.h"

/********************************************/
/* DEBUG functions first, normal ones later */
/********************************************/
#ifdef POOLDEBUG 

pool _pools = NULL;
int _poolblocks = 0;
int _poolstat = 0;

pool pool_new_d(char *zone)
{
    pool p;
    p = malloc(sizeof(_pool));
    p->size = p->debug = 0;
    p->stat = p->cstat = _poolstat;
    p->label = strdup(zone);
    p->next = NULL;
    p->block = (void *)_pools;
    _pools = p;
/*    fprintf(stderr,"_ _ _ _ / %X %s:%d \\ _ _ _ _\n",p,p->file,p->line);*/
    return p;
}

void *pmalloc_d(char *zone, pool p, int size)
{
    void *block = NULL;
    pool new;

    block = malloc(size);

    if(p == NULL) /* this is most likely an error that we got a NULL here! */
    {
        fprintf(stderr,"Memory Leak! [pmalloc received NULL pool, unable to track allocation]");
    }else{
        new = malloc(sizeof(_pool));
        new->size = size;
        new->block = block;
        new->next = p->next;
        new->label = strdup(zone);
        new->stat = p->cstat = _poolstat;
        p->size += size;
        _poolblocks += size;
        p->next = new;
        if(p->debug)
            fprintf(stderr,"_ _ _ _ / %X +%d [%s/%s] \\ _ _ _ _\n",p,size,p->label,zone);
    }

    return block;
}

char *pstrdup_d(char *zone, pool p, const char *src)
{
    char *ret;

    if(src == NULL)
        return NULL;

    ret = pmalloc_d(zone, p,strlen(src) + 1);
    strcpy(ret,src);

    return ret;
}

int pool_size(pool p)
{
    if(p == NULL) return 0;

    return p->size;
}

void pool_free(pool p)
{
    pool next, current;

    if(p == NULL)
        return;

    /* unlink p from the master list */
    if(p == _pools)
        _pools = (pool)(_pools->block);
    else
        for(current = _pools; current != NULL; current = (pool)(current->block))
            if(p == (pool)(current->block))
                current->block = (pool)(p->block);

    if(p->debug)
        fprintf(stderr,"_ _ _ _ \\ %X -%d [%s] / _ _ _ _\n",p,p->size,p->label);
    p->size = 0;
    p->block = NULL;
    current = p;
    while(current != NULL)
    {
        next = current->next;
        _poolblocks -= current->size;
        free(current->block);
        free(current->label);
        free(current);
        current = next;
    }
}

void pool_stat(int full)
{
    pool cur, cur2;

    fprintf(stderr,"_ _ _ _ > POOL %d < _ _ _ _\n",_poolblocks);
    for(cur = _pools; cur != NULL; cur = (pool)(cur->block))
    {
        if(cur->stat == _poolstat)
            fprintf(stderr,"%X %d (new)%s\n",cur,cur->size,cur->label);
        else if(full || cur->cstat == _poolstat)
            fprintf(stderr,"%X %d %s\n",cur,cur->size,cur->label);

        continue;
        if(cur->stat != _poolstat && cur->cstat != _poolstat) continue;
        if(cur->cstat)
            for(cur2 = cur->next; cur2 != NULL; cur2 = cur2->next)
                if(cur2->stat == _poolstat)
                    fprintf(stderr,"        +%d %s\n",cur2->size,cur2->label);
    }
    fprintf(stderr,"_ _ _ _ _ _ _ _ _ _ _ _ _ _ _\n");
    ++_poolstat;
}

void pool_label(pool p, char *label, int debug)
{
    if(p == NULL || label == NULL) return;

    free(p->label);
    p->label = strdup(label);    
    p->debug = debug;
}

/********************************************/
#else
/********************************************/

pool pool_new(void)
{
    pool p;
    p = malloc(sizeof(_pool));
    p->block = NULL;
    p->next = NULL;
    p->size = 0;
    return p;
}

void *pmalloc(pool p, int size)
{
    void *block = NULL;
    pool new;

    block = malloc(size);

    if(p == NULL) /* this is most likely an error that we got a NULL here! */
    {
        fprintf(stderr,"Memory Leak! [pmalloc received NULL pool, unable to track allocation]");
    }else{
        new = malloc(sizeof(_pool));
        new->size = size;
        new->block = block;
        new->next = p->next;
        p->size += size;
        p->next = new;
    }

    return block;
}

char *pstrdup(pool p, const char *src)
{
    char *ret;

    if(src == NULL)
        return NULL;

    ret = pmalloc(p,strlen(src) + 1);
    strcpy(ret,src);

    return ret;
}

int pool_size(pool p)
{
    if(p == NULL) return 0;

    return p->size;
}

void pool_free(pool p)
{
    pool next, current;

    current = p;
    while(current != NULL)
    {
        next = current->next;
        free(current->block);
        free(current);
        current = next;
    }
}

void pool_label(pool p, char *label, int debug)
{
    return;
}

void pool_stat(int full)
{
    return;
}

/********************************************/
#endif /* POOLDEBUG */
/********************************************/
